/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)41       1.79.1.25  src/avs/fs/mmfs/ts/kernext/gpl-linux/verdep.h, mmfs, avs_rgpfs32, rgpfs32s021a 2/17/10 09:12:46 */
#ifndef _h_verdep
#define _h_verdep

#if (LINUX_KERNEL_VERSION >= 2060000)
/************************************
 * Linux 2.6 functional differences *
 ************************************/
# include <linux/version.h>
# include <linux/buffer_head.h>
# include <linux/kdev_t.h>
# include <linux/namei.h>
# include <linux/mount.h>
# include <linux/page-flags.h>
# include <linux/security.h>
# include <linux/sched.h>
# include <linux/init.h>
# include <linux/fs.h>

# include <cxiTypes.h>
# include <cxiSystem.h>
# include <Logger-gpl.h>
# include <linux/moduleparam.h>


/* from 2.6.22, thread_info is moved to stack in  task_struct */
#if (LINUX_KERNEL_VERSION >= 2062200)
#define TASK_THREAD_INFO(task) ((struct thread_info *)(task)->stack)
#else 
#define TASK_THREAD_INFO(task) (task->thread_info)
#endif 

/* Distro-specific defines */
# if (defined(SUSE_LINUX) && defined(LINUX_KERNEL_VERSION) && LINUX_KERNEL_VERSION >= 2062700)
#    define SUSE_SLES11  
# elif  (defined(SUSE_LINUX) && SLE_VERSION_CODE >= 655872)
#    define SUSE_SLES10SP2
# endif
# if (defined(REDHAT_AS_LINUX) && defined(RHEL_RELEASE_CODE) && RHEL_RELEASE_CODE == 1282)
#    define REDHAT_RHEL52
#    define REDHAT_RHEL5
# endif
# if (defined(REDHAT_AS_LINUX) && defined(RHEL_RELEASE_CODE) && RHEL_RELEASE_CODE == 1283)
#    define REDHAT_RHEL53
#    define REDHAT_RHEL5
# endif

# if (defined(SUSE_LINUX) && defined(SLE_VERSION_CODE) && SLE_VERSION_CODE >= 656128)
#    define STATFS_USES_DENTRY
# endif

  /* Inode state indicators.
     SLES9 2.6.5-7.244+ kernels and SLES10 kernels require truncate_inode_pages
     to be called from the delete_inode sop and check for I_WILL_FREE flag, 
     RHEL4 kernels do not (thus far).  */
# if (defined(SUSE_LINUX) && \
     ((LINUX_KERNEL_VERSION == 2060507 && \
       SUSE_LINUX_UPDATE_SUBLEVEL >= 24400000000) || \
      LINUX_KERNEL_VERSION > 2060507)) || \
     (defined(REDHAT_AS_LINUX) && LINUX_KERNEL_VERSION > 2061800) || \
     (defined(KERNEL_ORG_LINUX) && LINUX_KERNEL_VERSION > 2061300)
     
#   define INODE_BEING_RELEASED (I_FREEING|I_CLEAR|I_WILL_FREE)
#   define INODE_INIT_INCOMPLETE (I_CLEAR|I_LOCK|I_NEW|I_WILL_FREE)
    /* required if file system defines delete_inode */
#   define MUST_TRUNCATE_INODE_PAGES 1
# else
#   define INODE_BEING_RELEASED (I_FREEING|I_CLEAR)
#   define INODE_INIT_INCOMPLETE (I_CLEAR|I_LOCK|I_NEW)
#   define MUST_TRUNCATE_INODE_PAGES 0
# endif

  /* Sort routine not provided in all kernels */
# if defined(REDHAT_LINUX) || (LINUX_KERNEL_VERSION >= 2061200)
#   define SORT(a,b,s,fn)   sort(a,b,s,fn,NULL)
# elif defined(SUSE_SLES9)
#   define SORT(a,b,s,fn)   qsort(a,b,s,fn)
# else
#   define SORT(a,b,s,fn)   do {} while (0)
/*  #error "No sort routine defined for this kernel" */
# endif

  /* Support for cluster-aware NFS locks in present in kernels > 2.6.22
     and backported to SLES10 SP2 and RHEL 5.2 */
# if defined(SUSE_SLES10SP2) || defined(REDHAT_RHEL52) || \
     (LINUX_KERNEL_VERSION >= 2062200)
#   define NFS_CLUSTER_LOCKS
# endif

  /* Support for cluster-aware leases */
# if defined(SUSE_SLES10SP2) || defined(REDHAT_RHEL53) || \
     (LINUX_KERNEL_VERSION >= 2062300)
#   define CLUSTER_LEASES
# endif

  /* Call to issue POSIX setlease */
# ifdef CLUSTER_LEASES
#   ifdef REDHAT_RHEL53
#     define POSIX_SETLEASE(fP, arg, flPP)  __setlease(fP, arg, flPP)
#   else
#     define POSIX_SETLEASE(fP, arg, flPP)  generic_setlease(fP, arg, flPP)
#   endif
# else
#     define POSIX_SETLEASE(fP, arg, flPP)  0
# endif

  /* Retrieve a new empty inode from the vfs */
# define NEW_INODE(SUPER)            new_inode(SUPER)
 
  /* Uncache a directory entry.  Must hold dcache lock. */
# define DENTRY_DROP(DP)  \
  {  \
    spin_lock(&(DP)->d_lock); \
    __d_drop(DP); \
    spin_unlock(&(DP)->d_lock); \
  }

  /* Uncache a directory entry.  Must hold *NO* locks. */
  /* __d_drop checks state of DCACHE_UNHASHED prior to drop. */
# define DENTRY_D_DROP(DP)  \
  {  \
    spin_lock(&dcache_lock); \
    spin_lock(&(DP)->d_lock); \
    __d_drop(DP); \
    spin_unlock(&(DP)->d_lock); \
    spin_unlock(&dcache_lock); \
  }

#define DENTRY_LOCK(DP) \
  { \
    spin_lock(&(DP)->d_lock); \
  }

#define DENTRY_UNLOCK(DP)\
  { \
    spin_unlock(&(DP)->d_lock); \
  }

  /* Check if directory is already uncached.  Must hold dcache lock. */
# define DENTRY_UNHASHED(DP)         d_unhashed(DP)

  /* Mark buffer in error and call iodone routine */
# define BUFHEAD_ERROR_IODONE(BH)  \
  {                                \
    clear_buffer_uptodate(BH);     \
    (BHP)->b_end_io((BH), 0);      \
  }


#if LINUX_KERNEL_VERSION >= 2061300 || \
   (LINUX_KERNEL_VERSION == 2060507 && defined(SUSE_LINUX) && SUSE_LINUX_UPDATE_SUBLEVEL >= 13900000000)
#define MY_INIT_LIST_HEAD(SINODES) INIT_LIST_HEAD(SINODES)
#else
#define MY_INIT_LIST_HEAD(SINODES) 
#endif

#if (LINUX_KERNEL_VERSION >= 2062400)
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
    MY_INIT_LIST_HEAD(&((SBP)->s_inodes));        \
  }

# define UNSET_SUPER_BLOCK(SBP)                   \
  {						  \
  }
#else
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
    security_sb_alloc(SBP);			  \
    MY_INIT_LIST_HEAD(&((SBP)->s_inodes));        \
  }

# define UNSET_SUPER_BLOCK(SBP)              \
  {						  \
    security_sb_free(SBP);			  \
  }
#endif

# define KSTATFS                     kstatfs
# define HAS_SOP_ALLOC_INODE         1
# define NODE_TOO_BIG_FOR_OSNODE(SZ) 0  /* false if HAS_SOP_ALLOC_INODE */

# define DAEMONIZE(A)                daemonize(A);
# define FOR_EACH_TASK(T)            for_each_process(T)
# define DO_EACH_THREAD(G,T)         do_each_thread(G,T)
# define WHILE_EACH_THREAD(G,T)      while_each_thread(G,T)
# define NEXT_TASK(T)                list_entry((T)->tasks.next, \
                                                 struct task_struct, tasks)

#if (LINUX_KERNEL_VERSION >= 2062500)
#define sb_entry(list)        list_entry((list), struct super_block, s_list)
#define TASK_PWD(tsP)         ((tsP)->fs ? &(tsP)->fs->pwd : NULL)
#define TASK_PWDMNT(tsP)      ((tsP)->fs ? (tsP)->fs->pwd.mnt : NULL)
#define TASK_ROOTMNT(tsP)     ((tsP)->fs ? (tsP)->fs->root.mnt : NULL)
#define NDP_TO_IP(ndP)        (ndP)->path.dentry->d_inode;
#define ND_PATH(nd)           (nd).path
#define EXP_DENTRY(expP)      (expP)->ex_path.dentry
#define PATH_PUT(ndP)         path_put(&(ndP)->path)
#else
#define TASK_PWD(tsP)         ((tsP)->fs ? (tsP)->fs->pwd : NULL)
#define TASK_PWDMNT(tsP)      ((tsP)->fs ? (tsP)->fs->pwdmnt : NULL)
#define TASK_ROOTMNT(tsP)     ((tsP)->fs ? (tsP)->fs->rootmnt : NULL)
#define NDP_TO_IP(ndP)        (ndP)->dentry->d_inode
#define ND_PATH(nd)           (nd)
#define EXP_DENTRY(expP)      (expP)->ex_dentry
#define PATH_PUT(ndP)         path_release(ndP)
#endif

#if (LINUX_KERNEL_VERSION >= 2062400)
# define PROCESS_GROUP(T)            task_pgrp_nr(T)
#else
# define PROCESS_GROUP(T)            process_group(T)
#endif
# define REVOKE_LEASE(INODE,MODE)    break_lease(INODE,MODE)
# define KDEV_INT(KDEV)              ((unsigned int)(KDEV))
# define PAGE_UP_TO_DATE(PAGE)       PageUptodate(PAGE)
# define PAGE_UNLOCK(PAGE)           unlock_page(PAGE)
# define HAS_IOP_REVALIDATE          0  /* uses getattr inode op instead */
# define HAS_SOP_READ_INODE2         0  
# define SBLOCK_PRIVATE(SBP)         (SBP)->s_fs_info
# define SBLOCK_BDEVNAME(SBP,BNAME)  bdevname((SBP)->s_bdev,(BNAME))

# define DCACHE_DFLAGS_DISCONNECTED  DCACHE_DISCONNECTED
# define POSIX_UNBLOCK_LOCK(FLP)     posix_unblock_lock((FLP)->fl_file,(FLP))

#if defined(NFS_CLUSTER_LOCKS) && !defined(REDHAT_RHEL52)
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2), NULL)
# define posix_lock_file_conf        posix_lock_file
#else
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2))
#endif
# define SPIN_LOCK_IRQ(LOCK,FLAGS)   spin_lock_irqsave(LOCK,FLAGS)
# define SPIN_UNLOCK_IRQ(LOCK,FLAGS) spin_unlock_irqrestore(LOCK,FLAGS)

# define GET_INODETIME_SEC(ITIME)  (ITIME).tv_sec
# define CXITIME_TO_INODETIME(CXITIME, ITIME)   \
  {                                             \
    (ITIME).tv_sec = (CXITIME).tv_sec;          \
    (ITIME).tv_nsec = (CXITIME).tv_nsec;        \
  }
# define CXITIME_FROM_INODETIME(CXITIME, ITIME) \
  {                                             \
    (CXITIME).tv_sec = (ITIME).tv_sec;          \
    (CXITIME).tv_nsec = (ITIME).tv_nsec;        \
  }

#ifdef MODULE
  /* Module function interfaces for 2.6 kernel */
# define MY_INIT_FUNCTION()           static int __init myModuleInit(void)
# define MY_EXIT_FUNCTION()           static void __exit myModuleExit(void)
# define MY_MODULE_INCREMENT()        try_module_get(THIS_MODULE)
# define MY_MODULE_DECREMENT()        module_put(THIS_MODULE) 

# define MODULE_INCREMENT(MOD)        try_module_get(MOD)
# define MODULE_DECREMENT(MOD)        module_put(MOD)
# define MODULE_USE_COUNT()           module_refcount(THIS_MODULE)

# define DEFINE_MODULE_INIT()         module_init(myModuleInit)
# define DEFINE_MODULE_EXIT()         module_exit(myModuleExit)
#endif /* MODULE */

/* module_param_string argument setting the permissions on
   prog_path and mmfsd_path, in /sys/module/mmfslinux/parameters */
#define PERM 0664

#define SETUP_MODULE_PATH_PARMS \
  char progPath[20] = "/usr/lpp/mmfs/bin/"; \
  char *prog_path = progPath; \
  char mmfsdPath[20] = "/usr/lpp/mmfs/bin/"; \
  char *mmfsd_path = mmfsdPath; \
  module_param_string(mmfsd_path, mmfsdPath, sizeof(mmfsdPath),PERM);\
  module_param_string(prog_path, progPath, sizeof(progPath),PERM);

  /* pte functions */
# define PTE_OFFSET(PMDP,VADDR)       pte_offset_map((PMDP),(VADDR))
# define PTE_UNMAP(PTEP)              pte_unmap(PTEP);

  /* mmap functions */
# define ACQUIRE_MMAP_SEM(SEM)        down_write(SEM)
# define RELEASE_MMAP_SEM(SEM)        up_write(SEM)

# define FILEMAP_FDATASYNC(RC, M)     RC = filemap_fdatawrite(M)
# define FILEMAP_FDATAWAIT(RC, M)     RC = filemap_fdatawait(M)

# define MOD_PGPGIN(amount)           add_page_state(pgpgin, (amount))
# define MOD_PGPGOUT(amount)          add_page_state(pgpgout, (amount))

#if (LINUX_KERNEL_VERSION >= 2060700)
#define PUTNAME __putname
#else
#define PUTNAME putname
#endif

#ifdef GPFS_ARCH_IA64
   /* user addresses are below region 5 (while PAGE_OFFSET is at region 7) */
#define GPFS_KERNEL_OFFSET 0xa000000000000000UL
#elif GPFS_ARCH_I386
#define GPFS_KERNEL_OFFSET (UIntPtr) PAGE_OFFSET 
#else
#define GPFS_KERNEL_OFFSET (UIntPtr) PAGE_OFFSET
#endif /* GPFS_ARCH_IA64 */

#define EXEC_HELPER(PATH,AV,EV,WAIT) call_usermodehelper(PATH,AV,EV,WAIT)
#define USING_MOUNT_HELPER() false

#if LINUX_KERNEL_VERSION >= 2060600 || \
   (LINUX_KERNEL_VERSION == 2060507 && defined(SUSE_LINUX) && SUSE_LINUX_UPDATE_SUBLEVEL >= 13900000000)
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) unmap_shared_mapping_range(MAPPING, START, STOP);
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = true;
#elif defined(MMAP_LINUX_PATCH)
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) invalidate_mappings(MAPPING, START);
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = true;
#else
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) 
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = false;
#endif

#if LINUX_KERNEL_VERSION >= 2060900
#define VFS_FOLLOW_LINK(RC, ND, BUF) nd_set_link(ND, BUF); BUF = NULL
#else
#define VFS_FOLLOW_LINK(RC, ND, BUF) RC = vfs_follow_link(ND, BUF)
#endif

#ifdef SUSE_SLES10SP2
#define VFS_MKDIR(vfsP, diP, dP, mode)        vfs_mkdir(diP, dP, vfsP, mode)
#define VFS_MKNOD(vfsP, diP, dP, mode, dev)   vfs_mknod(diP, dP, vfsP, mode, dev)
#define VFS_UNLINK(vfsP, diP, dP)             vfs_unlink(diP, dP, vfsP)
#define VFS_RMDIR(vfsP, diP, dP)              vfs_rmdir(diP, dP, vfsP)
#define VFS_LINK(vfsP, sdP, diP, tdP)         vfs_link(sdP, vfsP, diP, tdP, vfsP)
#define VFS_RENAME(vfsP, sdiP, sdP, tdiP, tdP) vfs_rename(sdiP, sdP, vfsP, tdiP, tdP, vfsP);
#elif defined SUSE_SLES11
#define VFS_MKDIR(vfsP, diP, dP, mode)        vfs_mkdir(diP, dP, vfsP, mode)
#define VFS_MKNOD(vfsP, diP, dP, mode, dev)   vfs_mknod(diP, dP, vfsP, mode, dev)
#define VFS_UNLINK(vfsP, diP, dP)             vfs_unlink(diP, dP, vfsP)
#define VFS_RMDIR(vfsP, diP, dP)              vfs_rmdir(diP, dP, vfsP)
#define VFS_LINK(vfsP, sdP, diP, tdP)         vfs_link(sdP, vfsP, diP, tdP, vfsP)
#define VFS_RENAME(vfsP, sdiP, sdP, tdiP, tdP) vfs_rename(sdiP, sdP, vfsP, tdiP, tdP, vfsP);
#else
#define VFS_MKDIR(vfsP, diP, dP, mode)        vfs_mkdir(diP, dP, mode)
#define VFS_MKNOD(vfsP, diP, dP, mode, dev)   vfs_mknod(diP, dP, mode, dev)
#define VFS_UNLINK(vfsP, diP, dP)             vfs_unlink(diP, dP)
#define VFS_RMDIR(vfsP, diP, dP)              vfs_rmdir(diP, dP)
#define VFS_LINK(vfsP, sdP, diP, tdP)         vfs_link(sdP, diP, tdP)
#define VFS_RENAME(vfsP, sdiP, sdP, tdiP, tdP) vfs_rename(sdiP, sdP, tdiP, tdP);
#endif

#if (!defined(SUSE_SLES11) && LINUX_KERNEL_VERSION >= 2062700) || defined(SUSE_SLES9)
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, nameP);
#elif defined(SUSE_SLES10SP2) 
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, vfsP, nameP, 0);
#elif defined(SUSE_SLES11)
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, vfsP, nameP);
#else
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, nameP, 0);
#endif

#if LINUX_KERNEL_VERSION >= 2061000
#define MY_RLIM_CUR(WHAT) current->signal->rlim[WHAT].rlim_cur
#else
#define MY_RLIM_CUR(WHAT) current->rlim[WHAT].rlim_cur
#endif

#if LINUX_KERNEL_VERSION >= 2061000
#define GENERIC_PERMISSION gpfs_i_permission_noacl
#else
#define GENERIC_PERMISSION vfs_permission
#endif

#if LINUX_KERNEL_VERSION >= 2061600
#define d_child d_u.d_child
#endif

#if LINUX_KERNEL_VERSION >= 2060900
#define HAS_IOP_PUT_LINK
#endif

#if LINUX_KERNEL_VERSION > 2061800 || defined(REDHAT_RHEL52)
#define PRVINODE i_private
#else
#define PRVINODE u.generic_ip
#endif

#define cxiDevMajor(D)       ((UInt32)((D) >> 20))
#define cxiDevMinor(D)       ((UInt32)((D) & 0xfffff))
#define cxiDevToKernelDev(D) new_decode_dev(D)
#define cxiDev32ToDev(D) \
  (cxiDev_t)((cxiDev32Major(D) << 8) | \
             (cxiDev32Minor(D) & 0xFF) | \
             ((cxiDev32Minor(D) & ~0xFF) << 12))

#if LINUX_KERNEL_VERSION >= 2062000
#define SLAB_KERNEL GFP_KERNEL
#endif

#if LINUX_KERNEL_VERSION < 2061600
#define INODE_LOCK(iP)    down(&(iP)->i_sem)
#define INODE_UNLOCK(iP)  up(&(iP)->i_sem)
#else
#define INODE_LOCK(iP)    mutex_lock(&(iP)->i_mutex)
#define INODE_UNLOCK(iP)  mutex_unlock(&(iP)->i_mutex)
#endif

#if LINUX_KERNEL_VERSION < 2061800
#define HAS_TASKLIST_LOCK
#define TASK_READ_LOCK read_lock(&tasklist_lock);
#define TASK_READ_UNLOCK read_unlock(&tasklist_lock);
#else
#include <linux/rcupdate.h>
#endif


#if LINUX_KERNEL_VERSION >= 2061900 || \
    (defined(REDHAT_AS_LINUX) && LINUX_KERNEL_VERSION >= 2061800)
#define INODE_USES_BLKBITS
#endif 

#if defined(INODE_USES_BLKBITS)
#define SET_INODE_BLOCKSIZE(inodeP, blocksize) \
  (inodeP)->i_blkbits = myblksize_bits((unsigned int)(blocksize));
#define GET_INODE_BLOCKSIZE(inodeP) (1 << (inodeP)->i_blkbits)
#else
#define SET_INODE_BLOCKSIZE(inodeP, blocksize) \
  (inodeP)->i_blksize = (blocksize);
#define GET_INODE_BLOCKSIZE(inodeP) ((inodeP)->i_blksize)
#endif

#if LINUX_KERNEL_VERSION < 2061700
#define HAS_PIDTYPE_TGID
#endif

#if LINUX_KERNEL_VERSION >= 2062000
#define KMEM_CACHE_T struct kmem_cache
#else
#define KMEM_CACHE_T kmem_cache_t
#endif

#define LOCK_FILES_STRUCT spin_lock(&current->files->file_lock)
#define UNLOCK_FILES_STRUCT spin_unlock(&current->files->file_lock)

#else /* < 2060000 */
/************************************
 * Linux 2.4 functional differences *
 ********************************** */
# include <linux/locks.h>
# include <linux/kernel_stat.h>

# include <cxiTypes.h>
# include <cxiSystem.h>
# include <Logger-gpl.h>

/* Inode state indicators */
#define INODE_BEING_RELEASED (I_FREEING|I_CLEAR)
#define INODE_IN_CACHE (I_FREEING|I_CLEAR|I_LOCK)
#define MUST_TRUNCATE_INODE_PAGES 0

#define PRVINODE u.generic_ip

#define GENERIC_PERMISSION vfs_permission

#define SORT(a,b,s,fn)   do {} while (0)

#if defined(REDHAT_AS_LINUX) || defined(SUSE_LINUX) ||\
    (defined(REDHAT_LINUX) && LINUX_KERNEL_VERSION >= 2041814)
  /* Retrieve a new empty inode from the vfs */
# define NEW_INODE(SUPER)            new_inode(SUPER)
#else
# define NEW_INODE(SUPER)            get_empty_inode()
#endif

#define INODE_LOCK(iP)    down(&(iP)->i_sem)
#define INODE_UNLOCK(iP)  up(&(iP)->i_sem)

  /* Check if directory is already uncached.  Must hold dcache lock. */
# define DENTRY_UNHASHED(DP)         list_empty(&(DP)->d_hash)

  /* Uncache a directory entry.  Assumes dcache lock is held */
# define DENTRY_DROP(DP) \
  { \
    if (!DENTRY_UNHASHED((DP))) \
      list_del_init(&(DP)->d_hash); \
  }

  /* Uncache a directory entry.  Must hold *NO* locks. */
# define DENTRY_D_DROP(DP) \
  { \
    spin_lock(&dcache_lock); \
    list_del_init(&(DP)->d_hash); \
    spin_unlock(&dcache_lock); \
  }

#define DENTRY_LOCK(DP) \
  { \
    spin_lock(&dcache_lock); \
  }

#define DENTRY_UNLOCK(DP)\
  { \
    spin_unlock(&dcache_lock); \
  }

  /* Mark buffer in error and call iodone routine */
# define BUFHEAD_ERROR_IODONE(BH)    buffer_IO_error(BH)

#if LINUX_KERNEL_VERSION >= 2041800
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
  }
#else
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
  }
#endif
# define UNSET_SUPER_BLOCK(SBP)              

#if ((defined(REDHAT_LINUX) && (LINUX_DISTRIBUTION_LEVEL >= 90)) || \
     (defined(REDHAT_AS_LINUX) && (LINUX_KERNEL_VERSION >= 2042000)))
# define HAS_01_SCHED_THREADING
#endif

#ifdef HAS_01_SCHED_THREADING
# define NEXT_TASK(T)                list_entry((T)->tasks.next, \
                                                struct task_struct, tasks)
# define FOR_EACH_TASK(T)            for ((T) = &init_task;        \
                                          ((T) = NEXT_TASK(T)) != &init_task; )
#else
# define FOR_EACH_TASK(T)            for_each_task(T)
# define NEXT_TASK(T)                (T)->next_task
#endif
# define DO_EACH_THREAD(G,T)         FOR_EACH_TASK(T)
# define WHILE_EACH_THREAD(G,T)      
# define PROCESS_GROUP(T)            (T)->pgrp

# define KSTATFS                     statfs
# define HAS_SOP_ALLOC_INODE         0
# define NODE_TOO_BIG_FOR_OSNODE(SZ) \
    ((sizeof(((struct inode *)0)->u) - sizeof(void *)) < (SZ))

#define TASK_PWD(tsP)                ((tsP)->fs ? (tsP)->fs->pwd : NULL)
#define TASK_PWDMNT(tsP)             ((tsP)->fs ? (tsP)->fs->pwdmnt : NULL)
#define TASK_ROOTMNT(tsP)            ((tsP)->fs ? (tsP)->fs->rootmnt : NULL)
#define NDP_TO_IP(ndP)               (ndP)->dentry->d_inode
#define ND_PATH(nd)                  (nd)
#define EXP_DENTRY(expP)             (expP)->ex_dentry
#define PATH_PUT(ndP)                path_release(ndP)

# define DAEMONIZE(A)                do { daemonize(); reparent_to_init(); } \
                                     while (0)
# define REVOKE_LEASE(INODE,MODE)    get_lease(INODE,MODE)
# define KDEV_INT(KDEV)              (KDEV)
# define PAGE_UP_TO_DATE(PAGE)       Page_Uptodate(PAGE)
# define PAGE_UNLOCK(PAGE)           UnlockPage(PAGE)
# define HAS_SOP_READ_INODE2         1 /* read_inode2 super block op present */
# define HAS_IOP_REVALIDATE          1 /* revalidate inode op present */
# define SBLOCK_PRIVATE(SBP)         (SBP)->u.generic_sbp
# define SBLOCK_BDEVNAME(SBP,BNAME)  bdevname((SBP)->s_dev)
# define BDEVNAME_SIZE               0 /* char array unneeded on 2.4 */

# define DCACHE_DFLAGS_DISCONNECTED  DCACHE_NFSD_DISCONNECTED
# define POSIX_UNBLOCK_LOCK(FLP)     posix_unblock_lock(FLP)
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2),0)
# define SPIN_LOCK_IRQ(LOCK,FLAGS)   spin_lock_irqsave(LOCK,FLAGS)
# define SPIN_UNLOCK_IRQ(LOCK,FLAGS) spin_unlock_irqrestore(LOCK,FLAGS)

# define GET_INODETIME_SEC(ITIME)  (ITIME)
# define CXITIME_TO_INODETIME(CXITIME, ITIME)   \
  {                                             \
    (ITIME) = (CXITIME).tv_sec;                 \
  }
# define CXITIME_FROM_INODETIME(CXITIME, ITIME) \
  {                                             \
    (CXITIME).tv_sec = (ITIME);                 \
    (CXITIME).tv_nsec = 0;                      \
  }

#ifdef MODULE
  /* Module function interfaces for 2.4 kernel */
# define MY_INIT_FUNCTION()           int init_module(void)
# define MY_EXIT_FUNCTION()           void cleanup_module(void)
# define MY_MODULE_INCREMENT()        MOD_INC_USE_COUNT
# define MY_MODULE_DECREMENT()        if (MOD_IN_USE) \
                                        MOD_DEC_USE_COUNT

# define MODULE_INCREMENT(MOD)        __MOD_INC_USE_COUNT(MOD)
# define MODULE_DECREMENT(MOD)        if (__MOD_IN_USE(MOD)) \
                                        __MOD_DEC_USE_COUNT(MOD)
# define MODULE_USE_COUNT()           GET_USE_COUNT(THIS_MODULE)

# define DEFINE_MODULE_INIT()
# define DEFINE_MODULE_EXIT()
# define SETUP_MODULE_PATH_PARMS \
  char *prog_path = NULL; \
  char *mmfsd_path = NULL; \
  MODULE_PARM(prog_path,"s"); \
  MODULE_PARM(mmfsd_path,"s");

#else
   /* No ops if not building a module */
# define MY_INIT_FUNCTION()          int myInit(void)
# define MY_EXIT_FUNCTION()          void myExit(void)
# define MY_MODULE_INCREMENT()
# define MY_MODULE_DECREMENT()

# define MODULE_INCREMENT(MOD)
# define MODULE_DECREMENT(MOD)
# define MODULE_USE_COUNT()          (1)

# define DEFINE_MODULE_INIT()        module_init(myInit)
# define DEFINE_MODULE_EXIT()        module_exit(myExit)
# define SETUP_MODULE_PATH_PARMS \
  char *prog_path = NULL; \
  char *mmfsd_path = NULL; 
#endif /* !MODULE */


  /* pte functions */
#if defined(REDHAT_AS_LINUX) || \
    (defined(REDHAT_LINUX) && (LINUX_KERNEL_VERSION >= 2042006))
# define PTE_OFFSET(PMDP,VADDR)  pte_offset_map((PMDP),(VADDR))
# define PTE_UNMAP(PTEP)         pte_unmap(PTEP);
#else
# define PTE_OFFSET(PMDP,VADDR)  pte_offset(PMDP, VADDR)
# if defined(SUSE_LINUX) && defined(GPFS_ARCH_I386)
#   define PTE_UNMAP(PTEP)       pte_kunmap(PTEP);
# else
#   define PTE_UNMAP(PTEP)       /* nothing */
# endif
#endif /* pte functions */

#if LINUX_KERNEL_VERSION >= 2040312 || GPFS_ARCH_POWER
# define ACQUIRE_MMAP_SEM(SEM)   down_write(SEM)
# define RELEASE_MMAP_SEM(SEM)   up_write(SEM)

#else
# define ACQUIRE_MMAP_SEM(SEM)   down(SEM)
# define RELEASE_MMAP_SEM(SEM)   up(SEM)
#endif

#if LINUX_KERNEL_VERSION >= 2041800
  /* Earlier kernels do not have a return code */
# define FILEMAP_FDATASYNC(RC, M) RC = filemap_fdatasync(M)
# define FILEMAP_FDATAWAIT(RC, M) RC = filemap_fdatawait(M)
#else
# define FILEMAP_FDATASYNC(RC, M) do { filemap_fdatasync(M); RC = 0; } while (0)
# define FILEMAP_FDATAWAIT(RC, M) do { filemap_fdatawait(M); RC = 0; } while (0)
#endif

#if (defined(REDHAT_AS_LINUX)) && \
    (LINUX_KERNEL_VERSION >= 2042100)
# define MOD_PGPGIN(amount)  kstat_percpu[smp_processor_id()].pgpgin += amount
# define MOD_PGPGOUT(amount) kstat_percpu[smp_processor_id()].pgpgout+= amount
#else
# define MOD_PGPGIN(amount)  kstat.pgpgin += amount
# define MOD_PGPGOUT(amount) kstat.pgpgout += amount
#endif

#define PUTNAME putname

#ifdef GPFS_ARCH_IA64
 /* user addresses are below region 5 (while PAGE_OFFSET is at region 7) */
#define GPFS_KERNEL_OFFSET 0xa000000000000000UL
#elif GPFS_ARCH_I386
#define GPFS_KERNEL_OFFSET (UIntPtr) PAGE_OFFSET 
#else
#define GPFS_KERNEL_OFFSET (UIntPtr) PAGE_OFFSET
#endif /* GPFS_ARCH_IA64 */

#if defined(GPFS_ARCH_X86_64) 
#   define EXEC_HELPER(PATH,AV,EV,WAIT) exec_usermodehelper(PATH,AV,EV)
#   define USING_MOUNT_HELPER() false
#else
#   define EXEC_HELPER(PATH,AV,EV,WAIT) execve(PATH,AV,EV)
#   define USING_MOUNT_HELPER() true
#endif

#ifdef MMAP_LINUX_PATCH
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) invalidate_mappings(MAPPING, START);
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = true;
#else
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) 
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = false;
#endif

#define VFS_FOLLOW_LINK(RC, ND, BUF) RC = vfs_follow_link(ND, BUF)

#define MY_RLIM_CUR(WHAT) current->rlim[WHAT].rlim_cur

#define cxiDevMajor(D)       ((Int32)(((UInt32)(D) >> 8) & 0xFF))
#define cxiDevMinor(D)       ((Int32)((D) & 0xFF))

/* Extra defines not on other platforms to massage it back down
 * to a linux kernel specific 16 bits.
 */
#define cxiDevToKernelDev(D) ((cxiDevMajor(D) << 8) | (cxiDevMinor(D) & 0xFF))
#define cxiDev32ToDev(D) \
  (cxiDev_t)((cxiDev32Major(D) << 8) | (cxiDev32Minor(D) & 0xFF))

#define HAS_TASKLIST_LOCK
#define TASK_READ_LOCK read_lock(&tasklist_lock);
#define TASK_READ_UNLOCK read_unlock(&tasklist_lock);

#define SET_INODE_BLOCKSIZE(inodeP, blocksize) \
  (inodeP)->i_blksize = (blocksize);
#define GET_INODE_BLOCKSIZE(inodeP) ((inodeP)->i_blksize)


#define HAS_PIDTYPE_TGID

#define KMEM_CACHE_T kmem_cache_t

#define LOCK_FILES_STRUCT read_lock(&current->files->file_lock)
#define UNLOCK_FILES_STRUCT read_unlock(&current->files->file_lock)

#endif /* 2.4 kernel */

#if LINUX_KERNEL_VERSION >= 2062700
#define FIND_TASK_BY_PID find_task_by_vpid
typedef struct request_queue request_queue_t;
#define USER_PATH(pathname, p) user_path(pathname, &(p)->path)
#define USER_LPATH(pathname, p) user_lpath(pathname, &(p)->path)
#define ATOMIC_SET_FSCOUNT(p, value) atomic_long_set(p, value)
#else
#define FIND_TASK_BY_PID find_task_by_pid
#define USER_PATH(pathname, p) user_path_walk(pathname, p)
#define USER_LPATH(pathname, p) user_path_walk_link(pathname, p)
#define ATOMIC_SET_FSCOUNT(p, value) atomic_set(p, value)
#endif

#if LINUX_KERNEL_VERSION >= 2062600
#define SEM_COUNT(semP)         (semP)->count
#define SEM_WAIT_TASKLIST(semP) ((semP)->wait_list)
#else
#define SEM_COUNT(semP)         atomic_read(&(semP)->count)
#define SEM_WAIT_TASKLIST(semP) ((semP)->wait.task_list)
#endif
#if LINUX_KERNEL_VERSION >= 2062800
#define d_alloc_anon d_obtain_alias
#endif

#if LINUX_KERNEL_VERSION >= 2062900
#define CRED(ctP, fl)         (ctP)->cred->fl
#define CRED_GRP(ctP, fl)     (ctP)->cred->group_info->fl
#define DENTRY_OPEN(dP, mntP, flags) dentry_open(dP, mntP, flags, current->cred)
#else
#if LINUX_KERNEL_VERSION > 2060300
#define CRED_GRP(ctP, fl)     (ctP)->group_info->fl
#else
#define CRED_GRP(ctP, fl)     (ctP)->fl
#endif
#define CRED(ctP, fl)         (ctP)->fl
#define DENTRY_OPEN(dP, mntP, flags) dentry_open(dP, mntP, flags)
#endif

/* Translate to GPFS native disk format which used high and low 16 bits
 * for major/minor
 */
#define cxiDevToDev32(D) \
  (cxiDev32_t)((cxiDevMajor(D) << 16) | (cxiDevMinor(D) & 0xFFFF))

#define MAX_GPFS_KMALLOC (131072)

#endif /* _h_verdep */
